package geniusweb.protocol.session.learn;

import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.mock;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.List;

import org.junit.Before;
import org.junit.Test;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import geniusweb.deadline.Deadline;
import geniusweb.deadline.DeadlineRounds;
import geniusweb.protocol.session.SessionSettings;
import geniusweb.protocol.session.TeamInfo;
import geniusweb.references.Parameters;
import geniusweb.references.PartyRef;
import geniusweb.references.PartyWithParameters;
import geniusweb.references.PartyWithProfile;
import geniusweb.references.ProfileRef;
import tudelft.utilities.junit.GeneralTests;
import tudelft.utilities.logging.Reporter;

public class LearnSettingsTest extends GeneralTests<LearnSettings> {
	private final ObjectMapper jackson = new ObjectMapper();
	LearnSettings settings1, settings1a, settings2, settings3;
	private DeadlineRounds deadline;
	private List<TeamInfo> participants, participants2;
	private Parameters params;

	private final String serialized = "{\"LearnSettings\":{\"participants\":[{\"TeamInfo\":{\"parties\":[{\"party\":{\"partyref\":\"party1\",\"parameters\":{\"persistentstate\":\"6bb5f909-0079-43ac-a8ac-a31794391074\",\"negotiationdata\":[\"12b5f909-0079-43ac-a8ac-a31794391012\"]}},\"profile\":\"http://prof1\"}]}}],\"deadline\":{\"DeadlineRounds\":{\"rounds\":10,\"durationms\":10000}}}}";

	@Before
	public void before() throws URISyntaxException, JsonProcessingException {

		params = new Parameters();
		params = params.with("persistentstate",
				"6bb5f909-0079-43ac-a8ac-a31794391074");
		params = params.with("negotiationdata",
				Arrays.asList(("12b5f909-0079-43ac-a8ac-a31794391012")));

		// define participants
		participants = createTeam(params, 1);

		// participants2
//		ProfileRef profile2 = new ProfileRef("http://prof2");
//		PartyWithProfile partywithp2 = new PartyWithProfile(party1, profile2);
//		List<PartyWithProfile> team2pp = Arrays.asList(partywithp2);
//		TeamInfo team2 = new TeamInfo(team2pp);
		participants2 = createTeam(params, 2);

		deadline = new DeadlineRounds(10, 10000);
		Deadline deadline2 = new DeadlineRounds(20, 10000);

		settings1 = new LearnSettings(participants, deadline);
		settings1a = new LearnSettings(participants, deadline);
		settings2 = new LearnSettings(participants2, deadline);
		settings3 = new LearnSettings(participants, deadline2);

	}

	/**
	 * 
	 * @param params  the parameters for the party
	 * @param partynr the serialnr for the party, eg 1 or 2.
	 * @return a List of TeamInfo with just 1 team.
	 */
	private List<TeamInfo> createTeam(Parameters params, int partynr)
			throws URISyntaxException {
		PartyRef party1ref = new PartyRef("party" + partynr);
		PartyWithParameters party1 = new PartyWithParameters(party1ref, params);
		ProfileRef profile1 = new ProfileRef("http://prof" + partynr);
		PartyWithProfile partywithp1 = new PartyWithProfile(party1, profile1);
		List<PartyWithProfile> team1pp = Arrays.asList(partywithp1);
		TeamInfo team1 = new TeamInfo(team1pp);
		return Arrays.asList(team1);

	}

	@Override
	public List<List<LearnSettings>> getGeneralTestData() {
		return Arrays.asList(Arrays.asList(settings1, settings1a),
				Arrays.asList(settings2), Arrays.asList(settings3));
	}

	@Override
	public List<String> getGeneralTestStrings() {
		return Arrays.asList(
				"LearnSettings.TeamInfo.*PartyWithProfile.*PartyRef.*party1.*persistentstate=6b.*, negotiationdata=\\[12.*..*ProfileRef.*prof1.*,DeadlineRounds.*10,10000.*",
				"LearnSettings.TeamInfo.*PartyWithProfile.*PartyRef.*party2.*persistentstate=6b.*, negotiationdata=\\[12.*..*ProfileRef.*prof2.*,DeadlineRounds.*10,10000.*",
				"LearnSettings.TeamInfo.*PartyWithProfile.*PartyRef.*party1.*persistentstate=6b.*, negotiationdata=\\[12.*..*ProfileRef.*prof1.*,DeadlineRounds.*20,10000.*");
	}

	@SuppressWarnings("unused")
	@Test(expected = IllegalArgumentException.class)
	public void testNullParticipants() {
		new LearnSettings(null, deadline);
	}

	@SuppressWarnings("unused")
	@Test(expected = IllegalArgumentException.class)
	public void testNullDeadline() {
		new LearnSettings(participants, null);
	}

	@Test
	public void getMaxRuntimeTest() {
		assertEquals(10d, settings1.getMaxRunTime(), 0.00000001);
	}

	@SuppressWarnings("unused")
	@Test
	public void getProtocolTest() {
		// this will throw if not Learn protocol or another problem.
		Learn protocol = (Learn) settings1.getProtocol(mock(Reporter.class));
	}

	@SuppressWarnings("unused")
	@Test(expected = IllegalArgumentException.class)
	public void badPersistentTest() throws URISyntaxException {
		// hack. #1933.
		Parameters newparams = params.with("persistentstate",
				"\"notproperUUID\"");
		participants = createTeam(newparams, 1);

		new LearnSettings(participants, deadline);
	}

	@SuppressWarnings("unused")
	@Test(expected = IllegalArgumentException.class)
	public void persistentNotStringTest() throws URISyntaxException {
		Parameters newparams = params.with("persistentstate", 32);
		participants = createTeam(newparams, 1);

		new LearnSettings(participants, deadline);
	}

	@SuppressWarnings("unused")
	@Test
	public void persistentWithoutQuotesTest() throws URISyntaxException {
		Parameters newparams = params.with("persistentstate",
				"6bb5f909-0079-43ac-a8ac-a31794391074");
		participants = createTeam(newparams, 1);

		new LearnSettings(participants, deadline);
	}

	@SuppressWarnings("unused")
	@Test(expected = IllegalArgumentException.class)
	public void persistentExtraQuoteTest() throws URISyntaxException {
		Parameters newparams = params.with("persistentstate",
				"\"6bb5f909-0079-43ac-a8ac-a31794391074\"");
		participants = createTeam(newparams, 1);

		new LearnSettings(participants, deadline);
	}

	@SuppressWarnings("unused")
	@Test(expected = IllegalArgumentException.class)
	public void negotiationDataNotList() throws URISyntaxException {
		Parameters newparams = params.with("negotiationdata",
				"\"12b5f909-0079-43ac-a8ac-a31794391012\"");
		participants = createTeam(newparams, 1);
		new LearnSettings(participants, deadline);
	}

	@SuppressWarnings("unused")
	@Test(expected = IllegalArgumentException.class)
	public void negotiationDataBadContents() throws URISyntaxException {
		Parameters newparams = params.with("negotiationdata", "\"bad\"");
		participants = createTeam(newparams, 1);
		new LearnSettings(participants, deadline);
	}

	@SuppressWarnings("unused")
	@Test(expected = IllegalArgumentException.class)
	public void negotiationDataNotString() throws URISyntaxException {
		Parameters newparams = params.with("negotiationdata", 12);
		participants = createTeam(newparams, 1);
		new LearnSettings(participants, deadline);
	}

	@Test
	public void getTeamsTest() {
		assertEquals(participants, settings1.getTeams());
		assertEquals((Integer) 1, settings1.getTeamSize());
	}

	@Test
	public void getDeadlineTest() {
		assertEquals(deadline, settings1.getDeadline());
	}

	@Test
	public void withTeamTest() {
		SessionSettings newset = settings1.with(participants2.get(0));
		assertEquals(2, newset.getTeams().size());
	}

	@Test
	public void testDeserialize() throws IOException {
		SessionSettings obj = jackson.readValue(serialized,
				SessionSettings.class);
		System.out.println(obj);
		assertEquals(settings1, obj);
	}

	@Test
	public void testSerialize()
			throws JsonProcessingException, URISyntaxException {

		String string = jackson.writeValueAsString(settings1);
		System.out.println(string);
		assertEquals(serialized, string);
	}

}
