package geniusweb.protocol.session.saop;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import org.junit.Before;
import org.junit.Test;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import geniusweb.actions.Accept;
import geniusweb.actions.Action;
import geniusweb.actions.EndNegotiation;
import geniusweb.actions.Offer;
import geniusweb.actions.PartyId;
import geniusweb.deadline.Deadline;
import geniusweb.deadline.DeadlineTime;
import geniusweb.issuevalue.Bid;
import geniusweb.progress.Progress;
import geniusweb.progress.ProgressRounds;
import geniusweb.progress.ProgressTime;
import geniusweb.protocol.NegoState;
import geniusweb.protocol.ProtocolException;
import geniusweb.protocol.partyconnection.ProtocolToPartyConn;
import geniusweb.protocol.session.TeamInfo;
import geniusweb.references.Parameters;
import geniusweb.references.PartyRef;
import geniusweb.references.PartyWithParameters;
import geniusweb.references.PartyWithProfile;
import geniusweb.references.ProfileRef;
import tudelft.utilities.junit.GeneralTests;

public class SAOPStateTest extends GeneralTests<SAOPState> {
	private final long NOW = 1000;

	private PartyId party1 = new PartyId("party1");
	private PartyId party2 = new PartyId("party2");
	private PartyId party3 = new PartyId("party3");
	private List<Action> actions1 = new LinkedList<>();
	private Action action = mock(Action.class);
	private List<Action> actions2 = Arrays.asList(action);

	private ProtocolToPartyConn party1conn = mock(ProtocolToPartyConn.class);
	private ProtocolToPartyConn party2conn = mock(ProtocolToPartyConn.class);
	private ProtocolToPartyConn party3conn = mock(ProtocolToPartyConn.class);

	private List<PartyId> connections = Arrays.asList(party1, party2);
	private List<PartyId> connections2 = Arrays.asList(party1, party3);
	private List<PartyId> connections3 = Arrays.asList(party1, party2, party3);

	private Progress progresstime = mock(ProgressTime.class);
	private ProgressRounds progressrounds = mock(ProgressRounds.class);
	private ProgressRounds progressrounds1 = mock(ProgressRounds.class);
	private SAOPSettings settings = mock(SAOPSettings.class);
	private SAOPState state, state1, state1a, state2, state3, state4;

	private Bid bid1 = mock(Bid.class);
	private Bid otherbid = mock(Bid.class);
	private Accept accept1 = new Accept(party1, bid1);
	private Accept accept2 = new Accept(party2, bid1);
	private Accept acceptother = new Accept(party2, otherbid);
	private Offer offer1 = new Offer(party1, bid1);
	private EndNegotiation endNegotiation1 = new EndNegotiation(party1);

	private final ObjectMapper jackson = new ObjectMapper();

	private String serialized = "{\"SAOPState\":{\"actions\":[],\"connections\":[\"party1\",\"party2\"],\"progress\":{\"ProgressTime\":{\"duration\":1000000,\"start\":2000000}},\"settings\":{\"SAOPSettings\":{\"participants\":[{\"TeamInfo\":{\"parties\":[{\"party\":{\"partyref\":\"party1ref\",\"parameters\":{}},\"profile\":\"prof1\"}]}}],\"deadline\":{\"DeadlineTime\":{\"durationms\":123000}}}},\"partyprofiles\":{},\"error\":null}}";

	@Before
	public void before() throws URISyntaxException {
		when(party1conn.getParty()).thenReturn(party1);
		when(party2conn.getParty()).thenReturn(party2);
		when(party3conn.getParty()).thenReturn(party3);
		when(party1conn.toString()).thenReturn("conn1");
		when(party2conn.toString()).thenReturn("conn2");
		when(party3conn.toString()).thenReturn("conn3");
		when(action.toString()).thenReturn("act1");

		when(progressrounds.getTerminationTime())
				.thenReturn(new Date(System.currentTimeMillis() + 3600000l));
		when(progressrounds1.getTerminationTime())
				.thenReturn(new Date(System.currentTimeMillis() + 3600000l));
		when(progressrounds.advance()).thenReturn(progressrounds1);

		state1 = new SAOPState(actions1, connections, progresstime, settings,
				null, null);
		state1a = new SAOPState(actions1, connections, progresstime, settings,
				null, null);
		state2 = new SAOPState(actions2, connections, progresstime, settings,
				null, null);
		state3 = new SAOPState(actions1, connections2, progresstime, settings,
				null, null);
		state4 = new SAOPState(actions1, connections, progressrounds, settings,
				null, null);

		List<PartyId> conns = Arrays.asList(party1, party2);
		PartyWithParameters pwithp = new PartyWithParameters(
				new PartyRef("party1ref"), new Parameters());
		List<PartyWithProfile> teams = Arrays
				.asList(new PartyWithProfile(pwithp, new ProfileRef("prof1")));
		List<TeamInfo> participants = Arrays.asList(new TeamInfo(teams));
		Deadline deadline = new DeadlineTime(123000);
		SAOPSettings setts = new SAOPSettings(participants, deadline);
		state = new SAOPState(actions1, conns,
				new ProgressTime(1000000l, new Date(2000000l)), setts, null,
				null);

	}

	@Override
	public List<List<SAOPState>> getGeneralTestData() {
		return Arrays.asList(Arrays.asList(state1, state1a),
				Arrays.asList(state2), Arrays.asList(state3),
				Arrays.asList(state4));
	}

	@Override
	public List<String> getGeneralTestStrings() {
		return Arrays.asList(
				"SAOPState.*\\[\\].*party1.*party2.*ProgressTime.*",
				"SAOPState.*\\[act1\\].*party1.*party2.*ProgressTime.*",
				"SAOPState.*party1.*party3.*ProgressTime.*",
				"SAOPState.*party1.*party2.*ProgressRounds.*");
	}

	@Test
	public void constructWith0Connection() {
		new SAOPState(actions1, Arrays.asList(), progresstime, settings, null,
				null);
	}

	@Test
	public void constructWithNullConnection() {
		SAOPState state = new SAOPState(actions1, null, progresstime, settings,
				null, null);
		assertEquals(0, state.getConnections().size());
	}

	@Test
	public void testNextActor() {
		assertEquals(party1, state1.getNextActor());
	}

	@Test
	public void testNextActor2() {
		List<PartyId> conns = Arrays.asList(party2, party3);
		SAOPState state = new SAOPState(actions1, conns, progresstime, settings,
				null, null);
		assertEquals(party2, state.getNextActor());
	}

	@Test
	public void testNextActorAfterAction() {
		List<Action> actions = Arrays.asList(action);
		SAOPState state = new SAOPState(actions, connections, progresstime,
				settings, null, null);
		assertEquals(party2, state.getNextActor());
	}

	@Test
	public void testNextActorAfter2Actions() {
		List<Action> actions = Arrays.asList(action, action);
		SAOPState state = new SAOPState(actions, connections, progresstime,
				settings, null, null);
		assertEquals(party1, state.getNextActor());
	}

	@Test
	public void isFinalTest() {
		assertFalse(state1.isFinal(NOW));
		SAOPState state = new SAOPState(actions1, connections, progresstime,
				settings, null, mock(ProtocolException.class));
		assertTrue(state.isFinal(NOW));
	}

	@Test
	public void withExceptionTest() {
		SAOPState finalstate1 = state1.with(mock(ProtocolException.class));
		assertTrue(finalstate1.isFinal(NOW));
		SAOPState finalstate2 = finalstate1.with(mock(ProtocolException.class));
		assertTrue(finalstate2.isFinal(NOW));
	}

	@Test
	public void isOfferAgreementTest1() {
		List<Action> actions = Arrays.asList(offer1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertTrue(state.getAgreements().getMap().isEmpty());
	}

	@Test
	public void isOfferAcceptAgreementTest() {
		List<Action> actions = Arrays.asList(offer1, accept2);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertTrue(state.getAgreements().getMap().isEmpty());
	}

	@Test
	public void isOfferAcceptOfferAgreementTest() {
		List<Action> actions = Arrays.asList(offer1, accept1, offer1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertTrue(state.getAgreements().getMap().isEmpty());
	}

	@Test
	public void isOfferAcceptAcceptAgreementTest() {
		List<Action> actions = Arrays.asList(offer1, accept1, accept2);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertEquals(bid1, state.getAgreements().getMap().get(party1));
		assertTrue(state.isFinal(NOW));
	}

	@Test
	public void isAcceptOfferAcceptAgreementTest() {
		List<Action> actions = Arrays.asList(accept1, offer1, accept2);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertTrue(state.getAgreements().getMap().isEmpty());
	}

	@Test
	public void isOfferAcceptAcceptotherAgreementTest() {
		List<Action> actions = Arrays.asList(offer1, accept1, acceptother);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertTrue(state.getAgreements().getMap().isEmpty());
	}

	@Test(expected = IllegalArgumentException.class)
	public void withActionTest() {
		when(action.getActor()).thenReturn(party1);
		state1.with(party1, action);
	}

	@Test
	public void withConnectionTest() {
		SAOPState state = state1.with(party3, (PartyWithProfile) null);
		assertEquals(3, state.getConnections().size());
	}

	@Test
	public void isEndNegotiationFinalTest() {
		List<Action> actions = Arrays.asList(endNegotiation1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertTrue(state.getAgreements().getMap().isEmpty());
		assertEquals(null, state.getError());
		assertTrue(state.isFinal(NOW));

	}

	@Test
	public void isProtocolErrorFinal() {
		SAOPState state = new SAOPState(actions1, connections3, progressrounds,
				settings, null,
				new ProtocolException("test error", new PartyId("test")));
		assertTrue(state.isFinal(NOW));
		assertNotNull(state.getError());
	}

	@Test(expected = IllegalArgumentException.class)
	public void RefuseImmediateAccept() {
		Accept nullaccept = new Accept(party1, null);
		state1.with(party1, nullaccept);
	}

	@Test(expected = IllegalArgumentException.class)
	public void RefuseNotMyTurn() {
		List<Action> actions = Arrays.asList(offer1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);

		state.with(party1, offer1);
	}

	@Test(expected = IllegalArgumentException.class)
	public void RefuseNullAccept() {
		List<Action> actions = Arrays.asList(offer1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);

		Accept nullaccept = new Accept(party2, null);
		state = state.with(party2, nullaccept);
		assertTrue(state.isFinal(NOW));
	}

	@Test(expected = IllegalArgumentException.class)
	public void RefuseNullAction() {
		List<Action> actions = Arrays.asList(offer1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);

		state = state.with(party2, (Action) null);
	}

	@Test
	public void checkGoodAccept() {
		List<Action> actions = Arrays.asList(offer1, accept2);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);

		Accept accept = new Accept(party3, bid1);
		state = state.with(party3, accept);
		assertTrue(state.isFinal(NOW));
		assertNull(state.getError());
		assertEquals(bid1, state.getAgreements().getMap().get(party1));

	}

	@Test
	public void checkGoodSecondAccept() {
		Bid bid2 = mock(Bid.class);
		List<Action> actions = Arrays.asList(offer1, new Offer(party2, bid2),
				new Accept(party3, bid2));
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);

		Accept accept = new Accept(party1, bid2);
		state = state.with(party1, accept);
		assertTrue(state.isFinal(NOW));
		assertNull(state.getError());
		assertEquals(bid2, state.getAgreements().getMap().get(party1));

	}

	@Test
	public void testDeserialize() throws IOException {
		NegoState obj = jackson.readValue(serialized, NegoState.class);
		System.out.println(obj);
		assertEquals(state, obj);
	}

	@Test
	public void testSerialize()
			throws JsonProcessingException, URISyntaxException {

		String string = jackson.writeValueAsString(state);
		System.out.println(string);
		assertEquals(serialized, string);
	}

}
